/* define_ext.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "databaseProgram.h"
#include "database.h"
#include "tableInfo.h"
#include "tableField.h"
#include "formDef.h"
#include "dataCell.h"
#include "create_tables.h"

#include <string>
#include <vector>

// This is a temporary program to populate the extension fields with
// values that suit me.  Eventually this will be done by the user via
// some clever gui.

struct ComboItem {
  int id;
  const char *label;
};

class DefineExtMain : public DatabaseProgram {
public:
  DefineExtMain();

  virtual void run();

  int combo_box(const ComboItem items[], int num_items,
		const string &description);

  void field(TableID table_id, int form_id, int left, int top,
	     const string &field_name, const string &label,
	     FormWidget::Type widget_type, vector<int> &slider_ids,
	     int xoptions = (GTK_FILL|GTK_EXPAND),
	     int yoptions = (GTK_FILL|GTK_EXPAND),
	     int xpadding = 0, int ypadding = 0, 
	     int uxsize = 0, int uysize = 0,
	     int type_assoc = -1);

  void cds_extension();
  void songs_extension();

  void define_ballroom_dances();

  void define_genre_query();
  void define_year_query();
  void define_title_query();
  void define_performer_query();
  void define_writer_query();
  void define_name_query();
  void define_general_query();

protected:
  virtual void handle_opts(const string &opt, const string &arg);

  virtual void help();
  virtual void describe_options();

  bool _include_ballroom;
  int _cds_form_id;
  int _songs_form_id;
  int _genre_combo_box_id;
};

ComboItem music_genres[] = {
  { 1, "Pop / Rock" },
  //  { 2, "Jazz / Blues / Swing" },
  { 16, "Jazz" },
  { 17, "Blues" },
  { 18, "Swing" },
  { 3, "New Age" },
  { 4, "Lounge / Vocal" },
  { 5, "Latin / International" },
  { 6, "Techno" },
  { 7, "Disco" },
  { 8, "Alternative" },
  { 9, "Country" },
  { 10, "Folk / Bluegrass / Cajun" },
  { 13, "Gospel" },
  { 11, "Classical" },
  { 12, "Soundtrack" },
  { 14, "Christmas" },
  { 15, "Specialty" },
};

static const int num_music_genres = sizeof(music_genres) / sizeof(ComboItem);

DefineExtMain::
DefineExtMain() : DatabaseProgram("define_ext") {
  define_opt("b", false);
  _include_ballroom = false;

  _cds_form_id = -1;
  _songs_form_id = -1;
  _genre_combo_box_id = -1;
}

void DefineExtMain::
run() {
  // First, make sure the database exists and has all the tables in
  // it.
  create_tables();

  // Now remove all the extension fields that may already have been
  // defined.  We'll redefine them all.
  music_db->sql_command
    ("DELETE FROM extension_fields");
  music_db->sql_command
    ("DELETE FROM forms");
  music_db->sql_command
    ("DELETE FROM default_forms");
  music_db->sql_command
    ("DELETE FROM form_widgets");
  music_db->sql_command
    ("DELETE FROM ratio_sliders");
  music_db->sql_command
    ("DELETE FROM form_relations");
  music_db->sql_command
    ("DELETE FROM form_queries");
  music_db->sql_command
    ("DELETE FROM combo_boxes");
  music_db->sql_command
    ("DELETE FROM combo_items");
  music_db->sql_command
    ("DELETE FROM combo_fields");

  // Define all the forms and their associated fields.
  cds_extension();
  songs_extension();

  define_genre_query();
  define_year_query();
  define_title_query();
  define_performer_query();
  define_writer_query();
  define_name_query();
  define_general_query();


  music_db->sql_command
    ("INSERT INTO default_forms (table_id, form_id) VALUES (" + 
     DataCell(TI_cds) + ", " + 
     DataCell(_cds_form_id) + ")");

  music_db->sql_command
    ("INSERT INTO default_forms (table_id, form_id) VALUES (" + 
     DataCell(TI_songs) + ", " + 
     DataCell(_songs_form_id) + ")");

  music_db->sql_command
    ("INSERT INTO form_relations (form_id, table_id, next_form) VALUES (" +
     DataCell(_cds_form_id) + ", " +
     DataCell(TI_songs) + ", " +
     DataCell(_songs_form_id) + ")");

  DatabaseProgram::run();
}

int DefineExtMain::
combo_box(const ComboItem items[], int num_items,
	  const string &description) {
  int combo_box_id = 
    music_db->get_next_sequence("combo_boxes_combo_box_id_seq");

  music_db->sql_command
    ("INSERT INTO combo_boxes (combo_box_id, description) VALUES (" +
     DataCell(combo_box_id) + ", " +
     DataCell(description) + ")");

  for (int i = 0; i < num_items; i++) {
    music_db->sql_command
      ("INSERT INTO combo_items (combo_item_id, label) VALUES (" +
       DataCell(items[i].id) + ", " +
       DataCell(items[i].label) + ")");

    music_db->sql_command
      ("INSERT INTO combo_fields (combo_box_id, combo_item_id, sequence) VALUES (" +
       DataCell(combo_box_id) + ", " +
       DataCell(items[i].id) + ", " +
       DataCell(i) + ")");
  }

  return combo_box_id;
}

void DefineExtMain::
field(TableID table_id, int form_id, int left, int top,
      const string &field_name, const string &label,
      FormWidget::Type widget_type, vector<int> &slider_ids,
      int xoptions, int yoptions, int xpadding, int ypadding, 
      int uxsize, int uysize, int type_assoc) {
  TableField field;
  field._field_type = FormWidget::get_db_type(widget_type);
  field._foreign_table = FormWidget::get_foreign_table(widget_type);
  if (field._field_type == DataCell::DC_bool) {
    field._default_value = DataCellBool(false);
  }
  field._field_name = field_name;
  field._label = label;
  field.insert_db(table_id);

  FormWidget widget;
  widget._table_id = table_id;
  widget._field_id = field._field_id;
  widget._widget_type = widget_type;
  widget._place_left = left;
  widget._place_right = left + 1;
  widget._place_top = top;
  widget._place_bottom = top + 1;
  widget._xoptions = xoptions;
  widget._yoptions = yoptions;
  widget._xpadding = xpadding;
  widget._ypadding = ypadding;
  widget._uxsize = uxsize;
  widget._uysize = uysize;
  widget._type_assoc = type_assoc;
  widget._label = label;
  if (widget_type != FormWidget::T_check_box) {
    widget._label += ":";
  }
  int widget_id = widget.insert_db(form_id);

  if (widget_type == FormWidget::T_check_box) {
    slider_ids.push_back(field._field_id);
  }
}

void DefineExtMain::
cds_extension() {
  _genre_combo_box_id = combo_box(music_genres, num_music_genres,
				 "Music Genres");

  FormDef form;
  form._name = "CD fields";
  form._num_rows = 1;
  form._num_columns = 2;
  form.record_db();
  _cds_form_id = form._form_id;

  vector<int> slider_ids;

  field(TI_cds_extension, form._form_id, 0, 0, "year", "Year", 
	FormWidget::T_year, slider_ids, 0, 0, 0, 0, 40, 0);

  field(TI_cds_extension, form._form_id, 1, 0, "genre", "Genre",
	FormWidget::T_combo_box, slider_ids, GTK_EXPAND, GTK_EXPAND, 
	0, 0, 0, 0, _genre_combo_box_id);
}


void DefineExtMain::
songs_extension() {
  if (_include_ballroom) {
    define_ballroom_dances();
  }
}


void DefineExtMain::
define_ballroom_dances() {
  FormDef form;
  form._name = "Ballroom Dances";
  form._y_size_pixels = 240;
  form._num_rows = 10;
  form._num_columns = 3;
  form._scrollbar = true;
  form.record_db();
  _songs_form_id = form._form_id;

  vector<int> slider_ids;

  field(TI_songs_extension, form._form_id, 0, 0, "bpm", "bpm", 
	FormWidget::T_number, slider_ids, 0, 0, 0, 4, 40, 0);
  
  field(TI_songs_extension, form._form_id, 0, 1, "mixer",
	"Mixer", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 2, "foxtrot",
	"Foxtrot", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 3, "socfoxtrot",
	"Society Foxtrot", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 4, "quickstep",
	"Quickstep", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 5, "slowwaltz",
	"Slow Waltz", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 6, "waltz",
	"Waltz", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 7, "vwaltz",
	"Viennese Waltz", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 8, "tango",
	"Tango", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 0, 9, "twostep",
	"Two-Step", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 1, "boogiewoogie",
	"Boogie Woogie", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 2, "swing1",
	"Single Swing", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 3, "swing2",
	"Double Swing", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 4, "swing3",
	"Triple Swing", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 5, "wcswing",
	"West Coast Swing", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 6, "hustle",
	"Hustle", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 7, "chacha",
	"Cha Cha", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 1, 8, "pasodoble",
	"Paso Doble", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 1, "rumba",
	"Rumba", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 2, "bolero",
	"Bolero", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 3, "merengue",
	"Merengue", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 4, "samba",
	"Samba", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 5, "mambo",
	"Mambo", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 6, "salsa",
	"Salsa", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 7, "cumbia",
	"Cumbia", FormWidget::T_check_box, slider_ids);
  field(TI_songs_extension, form._form_id, 2, 8, "lambada",
	"Lambada", FormWidget::T_check_box, slider_ids);

  // Now define the slider group.
  FormDef form2;
  form2._name = "Select Ballroom Dances";
  form2._num_rows = 1;
  form2._num_columns = 1;
  form2.record_db();
  int ballroom_dance_query_id = form2._form_id;

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_ratio_group;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._xoptions = (GTK_FILL | GTK_EXPAND);
  widget._yoptions = (GTK_FILL | GTK_EXPAND);

  int widget_id = widget.insert_db(form2._form_id);

  for (int s = 0; s < slider_ids.size(); s++) {
    music_db->sql_command
      ("INSERT INTO ratio_sliders (widget_id, table_id, field_id, sequence)"
       " VALUES (" +
       DataCell(widget_id) + ", " +
       DataCell(TI_songs_extension) + ", " +
       DataCell(slider_ids[s]) + ", " +
       DataCell(s) + ")");
  }

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(ballroom_dance_query_id) + ", " +
     DataCell(0.5) + ", " +
     DataCell("/Queries/Ballroom Dance") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_genre_query() {
  FormDef form;
  form._name = "Select CD's by genre";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command =
    "SELECT cds.cd_id, cds.performer, cds.title, cds_extension.genre "
    "FROM cds, cds_extension "
    "WHERE cds_extension.cd_id = cds.cd_id "
    "AND cds_extension.genre = $genre "
    "ORDER BY cds.title";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_combo_box;
  widget._type_assoc = _genre_combo_box_id;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "genre";
  widget._label = "Genre";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(1.0) + ", " +
     DataCell("/Queries/CD's by Genre") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_year_query() {
  FormDef form;
  form._name = "Select CD's by year";
  form._num_rows = 2;
  form._num_columns = 2;
  form._general_sql_command =
    "SELECT cds.cd_id, cds.title, cds.performer, cds_extension.year "
    "FROM cds_extension, cds "
    "WHERE cds_extension.cd_id = cds.cd_id "
    "AND cds_extension.year >= $from "
    "AND cds_extension.year <= $to "
    "ORDER BY cds.title";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_year;
  widget._place_left = 1;
  widget._place_right = 2;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._uxsize = 40;
  widget._internal_name = "from";
  widget.insert_db(form._form_id);

  widget._widget_id = -1;
  widget._place_top = 1;
  widget._place_bottom = 2;
  widget._internal_name = "to";
  widget.insert_db(form._form_id);

  widget._widget_id = -1;
  widget._widget_type = FormWidget::T_label;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "";
  widget._label = "From year";
  widget.insert_db(form._form_id);

  widget._widget_id = -1;
  widget._place_top = 1;
  widget._place_bottom = 2;
  widget._label = "To year";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(2.0) + ", " +
     DataCell("/Queries/CD's by Year") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_title_query() {
  FormDef form;
  form._name = "Select CD's by partial title";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command =
    "SELECT cds.cd_id, cds.performer, cds.title "
    "FROM cds "
    "WHERE lower(cds.title) like lower('%$title.u%')";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_string;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "title";
  widget._label = "Partial title (case insensitive)";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(2.5) + ", " +
     DataCell("/Queries/CD's by Title") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_performer_query() {
  FormDef form;
  form._name = "Select songs by performer";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command =
    "SELECT songs.song_id, songs.performer, songs.name,"
    " UPPER(" + music_db->concat(music_db->concat("artists.last_name",
						  "artists.first_name"),
				 "artists.group_name") + ") AS sort_artists"
    " FROM songs, artists"
    " WHERE songs.performer=$performer"
    " AND artists.artist_id=songs.performer"
    " ORDER BY sort_artists, songs.name";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_artist;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "performer";
  widget._label = "Performer";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(3.0) + ", " +
     DataCell("/Queries/Songs by Performer") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_writer_query() {
  FormDef form;
  form._name = "Select songs by writer";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command =
    "SELECT songs.song_id, songs.writer, songs.name,"
    " UPPER(" + music_db->concat(music_db->concat("artists.last_name",
						  "artists.first_name"),
				 "artists.group_name") + ") AS sort_artists"
    " FROM songs, artists"
    " WHERE songs.writer=$writer"
    " AND artists.artist_id=songs.writer"
    " ORDER BY sort_artists, songs.name";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_artist;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "writer";
  widget._label = "Writer";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(4.0) + ", " +
     DataCell("/Queries/Songs by Writer") + ", " +
     DataCell() + ")");
}
  

void DefineExtMain::
define_name_query() {
  FormDef form;
  form._name = "Select songs by partial name";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command =
    "SELECT songs.song_id, songs.performer, songs.name "
    "FROM songs "
    "WHERE lower(songs.name) like lower('%$name.u%')";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_string;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._internal_name = "name";
  widget._label = "Partial name (case insensitive)";
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(5.0) + ", " +
     DataCell("/Queries/Songs by Name") + ", " +
     DataCell() + ")");
}

void DefineExtMain::
define_general_query() {
  FormDef form;
  form._name = "General SQL command query";
  form._num_rows = 2;
  form._num_columns = 1;
  form._general_sql_command = "$sql.u";
  form.record_db();

  FormWidget widget;
  widget._table_id = TI_no_table;
  widget._field_id = -1;
  widget._widget_type = FormWidget::T_label;
  widget._place_left = 0;
  widget._place_right = 1;
  widget._place_top = 0;
  widget._place_bottom = 1;
  widget._label = "Enter a general SQL command:";
  widget._internal_name = "label";
  widget.insert_db(form._form_id);

  widget._widget_id = -1;
  widget._widget_type = FormWidget::T_text;
  widget._place_top = 1;
  widget._place_bottom = 2;
  widget._label = "";
  widget._internal_name = "sql";
  widget._xoptions = (GTK_FILL | GTK_EXPAND);
  widget._yoptions = (GTK_FILL | GTK_EXPAND);
  widget.insert_db(form._form_id);

  music_db->sql_command
    ("INSERT INTO form_queries (form_id, sequence, menu_item, accelerator) "
     "VALUES (" +
     DataCell(form._form_id) + ", " +
     DataCell(6.0) + ", " +
     DataCell("/Queries/General SQL query") + ", " +
     DataCell() + ")");
}
  


void DefineExtMain::
handle_opts(const string &opt, const string &arg) {
  if (opt == "b") {
    _include_ballroom = true;

  } else {
    DatabaseProgram::handle_opts(opt, arg);
  }
}

void DefineExtMain::
help() {
  cerr <<
    "\n"
    "define_ext is a temporary utility program to fill up the Paloma extension\n"
    "fields and query form definitions with reasonable defaults.\n";

  DatabaseProgram::help();
}

void DefineExtMain::
describe_options() {
  cerr <<
    "  -b         Include fields specialized for organizing songs into ballroom\n"
    "             dance categories.\n";
  DatabaseProgram::describe_options();
}

int
main(int argc, char *argv[]) {
  DefineExtMain m;
  m.command_line(argc, argv);
  m.init();
  m.run();
  return (0);
}
