/* postgresDatabase.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#ifdef USE_PGSQL

#include "postgresDatabase.h"
#include "postgresQueryResult.h"
#include "format.h"

#include <assert.h>

string PostgresDatabase::_user_args;


PostgresDatabase::
PostgresDatabase(const string &name, const string &host, int port,
		 const string &user, const string &passwd) {
  string connect = _user_args + " dbname='" + name + "'";

  if (!host.empty()) {
    connect += " host=" + host;
  }
  if (port > 0) {
    connect += " port=" + format_int(port);
  }
  if (!user.empty()) {
    connect += " user='" + user + "'";
  }
  if (!passwd.empty()) {
    connect += " password='" + passwd + "'";
  }

  _connection = PQconnectdb(connect.c_str());
  assert(_connection != NULL);

  if (PQstatus(_connection) == CONNECTION_BAD) {
    cerr << PQerrorMessage(_connection) << "\n"
	 << "Make sure Postgres is set up properly and you have created a database\n"
	 << "named '" << name << "'.\n";
    exit(1);
  }

  // We want to read all time values in Universal time; we'll do the
  // conversion locally.
  sql_command("SET TIME ZONE 'GMT'");

  check_db();
}



PostgresDatabase::
~PostgresDatabase() {
  PQfinish(_connection);
  _connection = NULL;
}

QueryResult PostgresDatabase::
sql_command(string command, bool return_failures) {
  PostgresQueryResult *pqr = new PostgresQueryResult(*this);

  if (command.empty()) {
    if (!return_failures) {
      cerr << "Attempt to execute an empty query!\n";
      abort();
    }
    return QueryResult(pqr);
  }

  if (command[command.length()-1] != ';') {
    command += ';';
  }
  //  cerr << command << "\n";

  PGresult *pg_result = PQexec(_connection, command.c_str());

  ExecStatusType status = (pg_result == NULL) ? PGRES_FATAL_ERROR :
    PQresultStatus(pg_result);

  switch (status) {
  case PGRES_EMPTY_QUERY:
  case PGRES_TUPLES_OK: 
  case PGRES_COMMAND_OK:
    pqr->set_pg_result(pg_result);
    break;

  case PGRES_BAD_RESPONSE:
  case PGRES_NONFATAL_ERROR:
    pqr->set_error_message(PQresultErrorMessage(pg_result));
    pqr->set_pg_result(pg_result);
    break;

  case PGRES_FATAL_ERROR:
    pqr->set_error_message(PQresultErrorMessage(pg_result));
    PQclear(pg_result);
    pg_result = NULL;
    break;

  case PGRES_COPY_OUT:
  case PGRES_COPY_IN:
  default:
    cerr << "Unexpected response to " << command << ": " << status << "\n";
    exit(1);
  }

  if (!pqr->success() && !return_failures) {
    cerr << "\nPostgreSQL error\n\n"
	 << pqr->get_error_message() << "\n"
	 << command << "\n";
    abort();
  }

  return QueryResult(pqr);
}

bool PostgresDatabase::
begin() {
  sql_command("BEGIN");
  return true;
}

bool PostgresDatabase::
commit() {
  sql_command("COMMIT");
  return true;
}

bool PostgresDatabase::
rollback() {
  sql_command("ROLLBACK");
  return true;
}

bool PostgresDatabase::
has_sequence(const string &name) {
  // This should be smarter--we should differentiate between sequences
  // and tables.
  return has_table(name);
}

bool PostgresDatabase::
create_sequence(const string &name) {
  return sql_command("CREATE SEQUENCE " + name).success();
}

int PostgresDatabase::
get_next_sequence(const string &name) {
  QueryResult qr = sql_command("SELECT NEXTVAL(" + DataCell(name) + ")");
  assert(qr.success());
  assert(qr.get_num_rows() == 1);
  assert(qr.get_num_columns() == 1);
  return qr.get_cell(0, 0).get_int();
}

string PostgresDatabase::
concat(const string &a, const string &b) {
  return "(" + a + " || " + b + ")";
}

string PostgresDatabase::
varchar_i(int size) {
  // Postgres just loves the TEXT type.
  return "TEXT";
}

bool PostgresDatabase::
has_table(const string &name) { 
  QueryResult qr = 
    sql_command("SELECT relname FROM pg_class "
		"WHERE relname=" + DataCell(name));
  if (!qr.success()) {
    cerr << "Unable to query table " << name << "\n";
    exit(1);
  }
  return (qr.get_num_rows() > 0);
}

void PostgresDatabase::
check_db() {
  if (PQstatus(_connection) == CONNECTION_OK) {
    return;
  }

  // The connection is lost for some reason.  Attempt to reestablish it.
  string error = PQerrorMessage(_connection);
  PQreset(_connection);

  if (PQstatus(_connection) == CONNECTION_OK) {
    // If that worked, swell.
    return;
  }

  // Otherwise, report the error that dropped the connection and quit.
  cerr << "No connection to database: " << error << "\n";
  exit(1);
}

/*
Table &PostgresDatabase::
define_table(const string &name) {
  Table *table = new Table(*this, name);
  
  // Ask Postgres about the table.
  PostgresQueryResult qr(*this);
  if (!sql_command("SELECT attnum,attname,atttypid FROM pg_class,pg_attribute "
		   "WHERE (pg_class.relname=" + DataCell(name) + ") and "
		   "(pg_class.oid=pg_attribute.attrelid) and "
		   "(attnum>0)", &qr)) {
    cerr << "Unable to query table " << name << "\n";

  } else if (qr.get_num_rows() == 0) {
    cerr << "Unknown table " << name << "\n";

  } else {
    assert(qr.get_num_columns() == 3);

    // Each row in the result represents a column in the table.
    int num_rows = qr.get_num_rows();
    for (int r = 0; r < num_rows; r++) {
      string col_name = qr.get_cell(r, 1).get_string();
      DataCell::Type type = get_type(qr.get_cell(r, 2).get_int());
      table->define_column(DataColumn(col_name, type, *table));
    }
  }

  _tables[name] = table;
  return *table;
}
*/

DataCell::Type PostgresDatabase::
get_type(int type_id) {
  Types::const_iterator ti;
  ti = _types.find(type_id);
  if (ti != _types.end()) {
    return (*ti).second;
  }

  // This is a newly-encountered type id.  Ask Postgres what it means.
  QueryResult qr =
    sql_command("SELECT typname FROM pg_type WHERE oid=" + 
		DataCell(type_id));
  if (!qr.success()) {
    cerr << "Unable to query type of " << type_id << "\n";
    abort();
  }

  PostgresQueryResult *pqr =
    (PostgresQueryResult *)qr.get_ptr();

  assert(qr.get_num_rows() == 1);
  assert(qr.get_num_columns() == 1);
  assert(pqr->get_cell_raw(0, 0) != NULL);
  const char *type_name_c_str = pqr->get_cell_raw(0, 0);
  assert(type_name_c_str != NULL);
  string type_name = type_name_c_str;

  DataCell::Type type = DataCell::DC_string;

  if (type_name == "bool") {
    type = DataCell::DC_bool;
  } else if (type_name.find("int") != string::npos) {
    type = DataCell::DC_int;
  } else if (type_name.find("float") != string::npos) {
    type = DataCell::DC_float;
  } else if (type_name.find("time") != string::npos) {
    type = DataCell::DC_timestamp;
  }

  _types[type_id] = type;
  return type;
}
      
void PostgresDatabase::
extra_args(const string &args) {
  _user_args += args + " ";
}

#endif  // USE_PGSQL
